//********************************************************
//
//	Linux Driver for CT-3000 Series Frame_Grabber 
//	   (CT-3000A,CT-3001,CT-3001RGB etc)
//
//	Copyright (C) 1999-2003 Cybertek CO.,
//
//	Released under the terms of the GPL.
//


#define MODULE
#include <linux/module.h>
#include <linux/version.h>

#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <asm/pgtable.h>
#include "ctdev.h"

#if LINUX_VERSION_CODE >= KERNEL_VER(2,4,9)
#include <linux/slab.h>
#else
#include <linux/malloc.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VER(2,4,0)
#include <sys/io.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,6)
#include <asm/uaccess.h>
#else
#include <asm/segment.h>
#include <linux/bios32.h>
#endif

static int ct_open( struct inode*, struct file*);
#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,6)
static int ct_release( struct inode*, struct file*);
#else
static void ct_release( struct inode*, struct file*);
#endif
static int ct_ioctl( struct inode*,struct file*,UINT,ULONG);
#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,6)
static int ct_mmap(struct file*, struct vm_area_struct*);
#else
static int ct_mmap(struct inode*, struct file*, struct vm_area_struct*);
#endif
void ct_vma_open(struct vm_area_struct*);
void ct_vma_close(struct vm_area_struct*);

static struct file_operations ct_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VER(2,4,0)
        THIS_MODULE,
#endif
	(loff_t)NULL,	       // lseek
	(ssize_t)NULL,	// read
	(ssize_t)NULL,	// write
	(int)NULL,		// readdir
	(UINT)NULL,		// poll
	ct_ioctl,		// ioctl
	ct_mmap,		// mmap
	ct_open,		// open
#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,6)
	(int)NULL,		// flush
#endif
	ct_release,		// release
	(int)NULL,		// fsync
	(int)NULL,		// fasync
	(int)NULL,		// check_media_change
//	(int)NULL,		// revalidate
	(int)NULL,		// lock
#if LINUX_VERSION_CODE >= KERNEL_VER(2,4,0)
        (ssize_t)NULL,     // readv
        (ssize_t)NULL      // writev
#endif
};

static struct vm_operations_struct ct_vm_ops = {
        ct_vma_open,    // open
        ct_vma_close,   // close
#if LINUX_VERSION_CODE >= KERNEL_VER(2,4,0)
        NULL,           // nopage
#else
        NULL,           // unmap
        NULL,           // protect
        NULL,           // sync
        NULL,           // advise
        NULL,           // nopage
        NULL,           // wppage
        NULL,           // swapout
        NULL,           // swapin
#endif
};

#if LINUX_VERSION_CODE >= KERNEL_VER(2,4,9)
// This gets rid of the "taint message"
MODULE_LICENSE("GPL");
#endif

int init_module(void)
{
  int result;
  
  result = register_chrdev(CT_MAJOR, "ctdev", &ct_fops);
  if(result < 0){
    printk("<1>register_chrdev failed.\n");
    return -EIO;
  }
  else printk("<1>CTDEV resident.\n");  
  return 0;
}

void cleanup_module(void)
{
  if(unregister_chrdev(CT_MAJOR, "ctdev") != 0)
    printk("<1>cleanup_module failed\n");
  else
    printk("<1>CTDEV removed.\n");
}


// open & close

static int ct_open( struct inode* inode, struct file* filp)
{
  int i;
  PUCHAR ptr;

  if(!pcibios_present()){
    printk("<1>pci bios not present.\n");
    return -EINVAL;
  }
  // secure private area
  filp->private_data = ptr = kmalloc(sizeof(DEVICE_TBL),GFP_KERNEL);
  if(filp->private_data == NULL){
    printk("<1>memory not available.\n");
    return -ENOMEM;
  }
  for(i=0;i<sizeof(DEVICE_TBL);i++) *ptr = 0;
  MOD_INC_USE_COUNT;
  return 0;   
}

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,6)
static int ct_release( struct inode* inode, struct file* filp)
#else
static void ct_release( struct inode* inode, struct file* filp)
#endif
{
  PDEVICE_TBL ptbl = filp->private_data;
  
  if(ptbl) kfree(ptbl);
  MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,6)
  return 0;
#endif
}

//
// The ioctl() implementation
//

static int ct_ioctl(struct inode *inode,struct file *filp,UINT cmd,ULONG arg)
{

  int i,err;
  ULONG tmp_data=0;
  PDEVICE_TBL ptbl = filp->private_data;
#if LINUX_VERSION_CODE >= KERNEL_VER(2,4,0)
  struct pci_dev *dev=NULL;
#else
  UCHAR bus,func;
#endif

  if(_IOC_TYPE(cmd) != CT_IOC_MAGIC) return -EINVAL;
  if(_IOC_NR(cmd) > CT_IOC_MAXNR) return -EINVAL; 
  err = verify_area(VERIFY_WRITE, (long *)arg, sizeof(long) * 4);
  if(err){
    printk("<1>ioctl reference error.\n");
    return -EINVAL;
  }

#if LINUX_VERSION_CODE > KERNEL_VER(2,1,3)
  copy_from_user(ptbl->parm,(void *)arg,sizeof(long) * 4l);
#else
  memcpy_fromfs(ptbl->parm,(void *)arg,sizeof(long) * 4l);
#endif
  

  switch(cmd){
    case (UINT)CT_IOCRESET :
      break;
    case (UINT)CT_PHYSMEM_MMAP :
      if(ptbl->parm[1]==(ULONG)-1){
#if LINUX_VERSION_CODE >= KERNEL_VER(2,4,0)
       if(dev != NULL) dev = NULL;
       for(i=0;i<=ptbl->parm[3];i++)
         dev = pci_find_device((UINT)CT_VENDOR_ID,(UINT)ptbl->parm[2],dev);
       if(dev == NULL) return -EINVAL;
	pci_read_config_dword(dev,PCI_BASE_ADDRESS_0,(u32 *)&(ptbl->PhyMemBase));
#else
       if(pcibios_find_device((UINT)CT_VENDOR_ID,(UINT)ptbl->parm[2],
	  (UINT)ptbl->parm[3],&bus,&func)) return -EINVAL;
	pcibios_read_config_dword(bus,func,PCI_BASE_ADDRESS_0,
          (u32 *)&(ptbl->PhyMemBase));
#endif
	ptbl->PhyMemBase &= PCI_BASE_ADDRESS_MEM_MASK;
       ptbl->MemSize  = CT_MEMORY_SIZE;
      }
      else{
        ptbl->PhyMemBase = ptbl->parm[1];
        ptbl->MemSize  = ptbl->parm[2];
      };
      tmp_data = ptbl->PhyMemBase;
      break;
    case (UINT)CT_OUTDW :
      outl(ptbl->parm[2],(UINT)ptbl->parm[1]);
      break;
    case (UINT)CT_INDW  :
      tmp_data = inl((UINT)ptbl->parm[1]);
      break;
    case (UINT)CT_GETVERSION :
      tmp_data = Driver_version;
      break;
    case (UINT)CT_GETPCICONF :
#if LINUX_VERSION_CODE >= KERNEL_VER(2,4,0)
       if(dev != NULL) dev = NULL;
       for(i=0;i<=ptbl->parm[2];i++)
         dev = pci_find_device((UINT)CT_VENDOR_ID,(UINT)ptbl->parm[1],dev);
       if(dev == NULL) return -EINVAL;
	pci_read_config_dword(dev,(UCHAR)ptbl->parm[3],(u32 *)&tmp_data);
#else
       if(pcibios_find_device((UINT)CT_VENDOR_ID,(UINT)ptbl->parm[1],
        (UINT)ptbl->parm[2],&bus,&func)) return -EINVAL;
      pcibios_read_config_dword(bus,func,(UCHAR)ptbl->parm[3],
        (u32 *)&tmp_data);
#endif
      break;
    case (UINT)CT_OUTW :
      outw((UINT)ptbl->parm[2],(UINT)ptbl->parm[1]);
      break;
    case (UINT)CT_INW :
      tmp_data = (ULONG)inw((UINT)ptbl->parm[1]);
      break;
    case (UINT)CT_OUTB :
      outb((UCHAR)ptbl->parm[2],(UINT)ptbl->parm[1]);
      break;
    case (UINT)CT_INB :
      tmp_data = (ULONG)inb((UINT)ptbl->parm[1]);
      break;
    case (UINT)CT_IOCHARDRESET :
      break;
    default :
      return -EINVAL;
  }
  put_user(tmp_data,(PULONG)arg);
  return 0;
}

//
//	The mmap() implementation
//

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,6)
static int ct_mmap(struct file * filp, struct vm_area_struct * vma)
#else
static int ct_mmap(struct inode *inode, struct file * filp, struct vm_area_struct * vma)
#endif
{
  
  PDEVICE_TBL ptbl = filp->private_data;
  ULONG msize,offset;


#if LINUX_VERSION_CODE < KERNEL_VER(2,4,0)
  if(vma->vm_offset & (PAGE_SIZE-1)) return -ENXIO; // need aligned offsets
  offset = vma->vm_offset;
#else
  offset = vma->vm_pgoff << PAGE_SHIFT;
#endif
  vma->vm_flags |= VM_RESERVED;
  msize = vma->vm_end - vma->vm_start;
  if(msize > ptbl->MemSize) msize = ptbl->MemSize;
  offset += ptbl->PhyMemBase;
  pgprot_val(vma->vm_page_prot) |= _PAGE_PCD; // caching disable 
  if(remap_page_range(vma->vm_start,offset,msize,vma->vm_page_prot))
    return -EAGAIN;
  vma->vm_ops = &ct_vm_ops;
  ct_vma_open(vma);
#if LINUX_VERSION_CODE < KERNEL_VER(2,1,6)
  vma->vm_inode = inode;
  inode->i_count++;
#endif
  return 0;
}

void ct_vma_open(struct vm_area_struct *area)  { MOD_INC_USE_COUNT; }

void ct_vma_close(struct vm_area_struct *area) { MOD_DEC_USE_COUNT; }

